@@ -241,8 +241,6 @@ class ModelCatalog:
             if action_space.dtype.name.startswith("int"):
                 low_ = np.min(action_space.low)
                 high_ = np.max(action_space.high)
-                assert np.all(action_space.low == low_)
-                assert np.all(action_space.high == high_)
                 dist_cls = TorchMultiCategorical if framework == "torch" \
                     else MultiCategorical
                 num_cats = int(np.product(action_space.shape))
@@ -738,8 +736,9 @@ class ModelCatalog:
             model_name (str): Name to register the model under.
             model_class (type): Python class of the model.
         """
-        if issubclass(model_class, tf.keras.Model):
-            deprecation_warning(old="register_custom_model", error=False)
+        if tf is not None:
+            if issubclass(model_class, tf.keras.Model):
+                deprecation_warning(old="register_custom_model", error=False)
         _global_registry.register(RLLIB_MODEL, model_name, model_class)
 
     @staticmethod
