@@ -30,6 +30,7 @@ from ray.autoscaler._private.util import DEBUG_AUTOSCALING_ERROR, \
     DEBUG_AUTOSCALING_STATUS
 from ray.internal.internal_api import memory_summary
 from ray.autoscaler._private.cli_logger import cli_logger, cf
+from distutils.dir_util import copy_tree
 
 logger = logging.getLogger(__name__)
 
@@ -451,6 +452,13 @@ def debug(address):
     type=str,
     help="Module path to the Python function that will be used to set up the "
     "environment for the worker process.")
+@click.option(
+    "--runtime-env-setup-hook",
+    hidden=True,
+    default=ray_constants.DEFAULT_RUNTIME_ENV_SETUP_HOOK,
+    type=str,
+    help="Module path to the Python function that will be used to set up the "
+    "runtime env in agent.")
 @add_click_options(logging_options)
 def start(node_ip_address, address, port, redis_password, redis_shard_ports,
           object_manager_port, node_manager_port, gcs_server_port,
@@ -462,7 +470,8 @@ def start(node_ip_address, address, port, redis_password, redis_shard_ports,
           no_redirect_output, plasma_store_socket_name, raylet_socket_name,
           temp_dir, system_config, lru_evict, enable_object_reconstruction,
           metrics_export_port, no_monitor, tracing_startup_hook,
-          worker_setup_hook, log_style, log_color, verbose):
+          worker_setup_hook, runtime_env_setup_hook, log_style, log_color,
+          verbose):
     """Start Ray processes manually on the local machine."""
     cli_logger.configure(log_style, log_color, verbose)
     if gcs_server_port and not head:
@@ -473,10 +482,6 @@ def start(node_ip_address, address, port, redis_password, redis_shard_ports,
     if node_ip_address is not None:
         node_ip_address = services.address_to_ip(node_ip_address)
 
-    redis_address = None
-    if address is not None:
-        (redis_address, redis_address_ip,
-         redis_address_port) = services.validate_redis_address(address)
     try:
         resources = json.loads(resources)
     except Exception:
@@ -484,7 +489,7 @@ def start(node_ip_address, address, port, redis_password, redis_shard_ports,
                          cf.bold("--resources"))
         cli_logger.abort(
             "Valid values look like this: `{}`",
-            cf.bold("--resources='\"CustomResource3\": 1, "
+            cf.bold("--resources='{\"CustomResource3\": 1, "
                     "\"CustomResource2\": 2}'"))
 
         raise Exception("Unable to parse the --resources argument using "
@@ -525,7 +530,8 @@ def start(node_ip_address, address, port, redis_password, redis_shard_ports,
         metrics_export_port=metrics_export_port,
         no_monitor=no_monitor,
         tracing_startup_hook=tracing_startup_hook,
-        worker_setup_hook=worker_setup_hook)
+        worker_setup_hook=worker_setup_hook,
+        runtime_env_setup_hook=runtime_env_setup_hook)
     if head:
         # Use default if port is none, allocate an available port if port is 0
         if port is None:
@@ -544,14 +550,36 @@ def start(node_ip_address, address, port, redis_password, redis_shard_ports,
             # not provided.
             num_redis_shards = len(redis_shard_ports)
 
-        if redis_address is not None:
-            cli_logger.abort(
-                "`{}` starts a new Redis server, `{}` should not be set.",
-                cf.bold("--head"), cf.bold("--address"))
-
-            raise Exception("If --head is passed in, a Redis server will be "
-                            "started, so a Redis address should not be "
-                            "provided.")
+        if address is not None:
+            cli_logger.print(
+                "Will use value of `{}` as remote Redis server address(es). "
+                "If the primary one is not reachable, we starts new one(s) "
+                "with `{}` in local.", cf.bold("--address"), cf.bold("--port"))
+            external_addresses = address.split(",")
+            reachable = False
+            try:
+                [primary_redis_ip, port] = external_addresses[0].split(":")
+                ray._private.services.wait_for_redis_to_start(
+                    primary_redis_ip, port, password=redis_password)
+                reachable = True
+            # We catch a generic Exception here in case someone later changes
+            # the type of the exception.
+            except Exception:
+                cli_logger.print(
+                    "The primary external redis server `{}` is not reachable. "
+                    "Will starts new one(s) with `{}` in local.",
+                    cf.bold(external_addresses[0]), cf.bold("--port"))
+            if reachable:
+                ray_params.update_if_absent(
+                    external_addresses=external_addresses)
+                if len(external_addresses) > 1:
+                    num_redis_shards = len(external_addresses) - 1
+                if redis_password == ray_constants.REDIS_DEFAULT_PASSWORD:
+                    cli_logger.warning(
+                        "`{}` should not be specified as empty string if "
+                        "external redis server(s) `{}` points to requires "
+                        "password.", cf.bold("--redis-password"),
+                        cf.bold("--address"))
 
         node_ip_address = services.get_node_ip_address()
 
@@ -579,10 +607,17 @@ def start(node_ip_address, address, port, redis_password, redis_shard_ports,
 
         node = ray.node.Node(
             ray_params, head=True, shutdown_at_exit=block, spawn_reaper=block)
+
         redis_address = node.redis_address
-        with open(
-                os.path.join(ray._private.utils.get_user_temp_dir(),
-                             "ray_current_cluster"), "w") as f:
+        if temp_dir is None:
+            # Default temp directory.
+            temp_dir = ray._private.utils.get_user_temp_dir()
+        # Using the user-supplied temp dir unblocks on-prem
+        # users who can't write to the default temp.
+        current_cluster_path = os.path.join(temp_dir, "ray_current_cluster")
+        # TODO: Consider using the custom temp_dir for this file across the
+        # code base. (https://github.com/ray-project/ray/issues/16458)
+        with open(current_cluster_path, "w") as f:
             print(redis_address, file=f)
 
         # this is a noop if new-style is not set, so the old logger calls
@@ -624,6 +659,10 @@ def start(node_ip_address, address, port, redis_password, redis_shard_ports,
             cli_logger.print(cf.bold("  ray stop"))
     else:
         # Start Ray on a non-head node.
+        redis_address = None
+        if address is not None:
+            (redis_address, redis_address_ip,
+             redis_address_port) = services.validate_redis_address(address)
         if not (port is None):
             cli_logger.abort("`{}` should not be specified without `{}`.",
                              cf.bold("--port"), cf.bold("--head"))
@@ -1494,13 +1533,21 @@ def status(address, redis_password):
     is_flag=True,
     default=True,
     help="Increase process information verbosity")
+@click.option(
+    "--tempfile",
+    "-T",
+    required=False,
+    type=str,
+    default=None,
+    help="Temporary file to use")
 def local_dump(stream: bool = False,
                output: Optional[str] = None,
                logs: bool = True,
                debug_state: bool = True,
                pip: bool = True,
                processes: bool = True,
-               processes_verbose: bool = False):
+               processes_verbose: bool = False,
+               tempfile: Optional[str] = None):
     """Collect local data and package into an archive.
 
     Usage:
@@ -1517,7 +1564,8 @@ def local_dump(stream: bool = False,
         debug_state=debug_state,
         pip=pip,
         processes=processes,
-        processes_verbose=processes_verbose)
+        processes_verbose=processes_verbose,
+        tempfile=tempfile)
 
 
 @cli.command()
@@ -1589,6 +1637,13 @@ def local_dump(stream: bool = False,
     is_flag=True,
     default=True,
     help="Increase process information verbosity")
+@click.option(
+    "--tempfile",
+    "-T",
+    required=False,
+    type=str,
+    default=None,
+    help="Temporary file to use")
 def cluster_dump(cluster_config_file: Optional[str] = None,
                  host: Optional[str] = None,
                  ssh_user: Optional[str] = None,
@@ -1600,7 +1655,8 @@ def cluster_dump(cluster_config_file: Optional[str] = None,
                  debug_state: bool = True,
                  pip: bool = True,
                  processes: bool = True,
-                 processes_verbose: bool = False):
+                 processes_verbose: bool = False,
+                 tempfile: Optional[str] = None):
     """Get log data from one or more nodes.
 
     Best used with Ray cluster configs:
@@ -1627,7 +1683,8 @@ def cluster_dump(cluster_config_file: Optional[str] = None,
         debug_state=debug_state,
         pip=pip,
         processes=processes,
-        processes_verbose=processes_verbose)
+        processes_verbose=processes_verbose,
+        tempfile=tempfile)
     if archive_path:
         click.echo(f"Created archive: {archive_path}")
     else:
@@ -1764,6 +1821,63 @@ def install_nightly(verbose, dryrun):
         subprocess.check_call(cmd)
 
 
+@cli.command()
+@click.option(
+    "--show-library-path",
+    "-show",
+    required=False,
+    is_flag=True,
+    help="Show the cpp include path and library path, if provided.")
+@click.option(
+    "--generate-bazel-project-template-to",
+    "-gen",
+    required=False,
+    type=str,
+    help="The directory to generate the bazel project template to,"
+    " if provided.")
+@add_click_options(logging_options)
+def cpp(show_library_path, generate_bazel_project_template_to, log_style,
+        log_color, verbose):
+    """Show the cpp library path and generate the bazel project template."""
+    if not show_library_path and not generate_bazel_project_template_to:
+        raise ValueError(
+            "Please input at least one option of '--show-library-path'"
+            " and '--generate-bazel-project-template-to'.")
+    cli_logger.configure(log_style, log_color, verbose)
+    raydir = os.path.abspath(os.path.dirname(ray.__file__))
+    cpp_templete_dir = os.path.join(raydir, "cpp/example")
+    include_dir = os.path.join(raydir, "cpp/include")
+    lib_dir = os.path.join(raydir, "cpp/lib")
+    if show_library_path:
+        cli_logger.print("Ray C++ include path {} ", cf.bold(f"{include_dir}"))
+        cli_logger.print("Ray C++ library path {} ", cf.bold(f"{lib_dir}"))
+    if generate_bazel_project_template_to:
+        if not os.path.isdir(generate_bazel_project_template_to):
+            cli_logger.abort(
+                "The provided directory "
+                f"{generate_bazel_project_template_to} doesn't exist.")
+        copy_tree(cpp_templete_dir, generate_bazel_project_template_to)
+        out_include_dir = os.path.join(generate_bazel_project_template_to,
+                                       "thirdparty/include")
+        if not os.path.exists(out_include_dir):
+            os.makedirs(out_include_dir)
+        copy_tree(include_dir, out_include_dir)
+        out_lib_dir = os.path.join(generate_bazel_project_template_to,
+                                   "thirdparty/lib")
+        if not os.path.exists(out_lib_dir):
+            os.makedirs(out_lib_dir)
+        copy_tree(lib_dir, out_lib_dir)
+
+        cli_logger.print(
+            "Project template generated to {}",
+            cf.bold(f"{os.path.abspath(generate_bazel_project_template_to)}"))
+        cli_logger.print("To build and run this template, run")
+        cli_logger.print(
+            cf.bold(
+                f"    cd {os.path.abspath(generate_bazel_project_template_to)}"
+                " && bazel run //:example"))
+
+
 def add_command_alias(command, name, hidden):
     new_command = copy.deepcopy(command)
     new_command.hidden = hidden
@@ -1796,6 +1910,7 @@ cli.add_command(cluster_dump)
 cli.add_command(global_gc)
 cli.add_command(timeline)
 cli.add_command(install_nightly)
+cli.add_command(cpp)
 
 try:
     from ray.serve.scripts import serve_cli
